@@ -1603,6 +1603,23 @@ quantize_fp8(const program& prog, const target& ptarget, const quantize_fp8_opti
          options.get_handle_ptr());
 }
 
+inline std::vector<std::string> get_onnx_operators()
+{
+    size_t size = 0;
+    call(&migraphx_get_onnx_operators_size, &size);
+    std::vector<std::string> result(size, "");
+
+    size_t index = 0;
+    for(auto& name : result)
+    {
+        char* name_op;
+        call(&migraphx_get_onnx_operator_name_at_index, &name_op, index);
+        name = name_op;
+        index++;
+    }
+    return result;
+}
+
 struct experimental_custom_op_base
 {
     experimental_custom_op_base()                                              = default;
